package com.samelement.iot_notification.service;

import android.app.Notification;
import android.app.NotificationChannel;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.appwidget.AppWidgetManager;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.SharedPreferences;
import android.graphics.Bitmap;
import android.graphics.Color;
import android.media.RingtoneManager;
import android.net.Uri;
import android.os.Build;
import android.os.Bundle;
import android.os.Handler;
import android.os.IBinder;
import android.os.Looper;
import android.text.Html;
import android.text.Spanned;
import android.util.Log;

import androidx.annotation.Nullable;
import androidx.core.app.NotificationCompat;
import androidx.localbroadcastmanager.content.LocalBroadcastManager;
import androidx.preference.PreferenceManager;
import androidx.work.Constraints;
import androidx.work.Data;
import androidx.work.ExistingWorkPolicy;
import androidx.work.NetworkType;
import androidx.work.OneTimeWorkRequest;
import androidx.work.WorkManager;

import com.google.gson.Gson;
import com.google.gson.JsonSyntaxException;
import com.samelement.iot_notification.R;
import com.samelement.iot_notification.analytic.Analytic;
import com.samelement.iot_notification.analytic.AnalyticResult;
import com.samelement.iot_notification.appwidget.AppWidgetType;
import com.samelement.iot_notification.appwidget.DemoAppWidget;
import com.samelement.iot_notification.appwidget.SimpleTextWidget;
import com.samelement.iot_notification.appwidget.db.AppWidgetObject;
import com.samelement.iot_notification.appwidget.db.WidgetReaderDbHelper;
import com.samelement.iot_notification.device.Device;
import com.samelement.iot_notification.device.DeviceResult;
import com.samelement.iot_notification.notification.MqttNotif;
import com.samelement.iot_notification.notification.MqttPayload;
import com.samelement.iot_notification.resource.AnalyticResource;
import com.samelement.iot_notification.resource.AnalyticsResourcesResult;
import com.samelement.iot_notification.resource.Resource;
import com.samelement.iot_notification.screen.AlarmActivity;
import com.samelement.iot_notification.screen.AlertActivity;
import com.samelement.iot_notification.screen.IntentKey;
import com.samelement.iot_notification.service.worker.GetAnalyticsResourcesWorker;
import com.samelement.iot_notification.service.worker.GetAnalyticsWorker;
import com.samelement.iot_notification.service.worker.GetDevicesWorker;
import com.samelement.iot_notification.service.worker.RestartServiceWorker;
import com.samelement.iot_notification.util.AnalyticIconHelper;
import com.samelement.iot_notification.util.EncryptUtil;
import com.samelement.iot_notification.util.FileUtil;
import com.samelement.iot_notification.util.FlutterUtil;
import com.samelement.iot_notification.util.NotificationHelper;
import com.samelement.iot_notification.util.StateIndicatorHelper;

import org.eclipse.paho.android.service.MqttAndroidClient;
import org.eclipse.paho.client.mqttv3.IMqttActionListener;
import org.eclipse.paho.client.mqttv3.IMqttDeliveryToken;
import org.eclipse.paho.client.mqttv3.IMqttToken;
import org.eclipse.paho.client.mqttv3.MqttCallbackExtended;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyManagementException;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

public class SamService extends Service {
    // error tag
    private final String TAG = getClass().getSimpleName();

    private int notificationId = 90909090;

    // MQTT Broker
    // because broker can be changed sometimes
    // and programmer don't want to deal with changing it every time.
    // broker is saved in SQLite database from flutter main application
    private String broker = "no-broker";
    private String brokerAddress = "no-host";
    private String brokerProtocol = "ssl"; // or tcp
    private int brokerPort = 8883; // or 1883
    private String cafile = "no-cafile";

    // we also save the username and password for this broker
    private String username = "invalid-token";
    private String password = "jwt";
    private String mqttClientIdentifier = UUID.randomUUID().toString();

    // MQTT Client and its connection options
    private MqttAndroidClient mqttAndroidClient;

    // list of topics and its QOS s
    private final List<SubscribeObj> subscribeObjList = Collections.synchronizedList(new ArrayList<SubscribeObj>());

    private SharedPreferences prefs;
    private WidgetReaderDbHelper dbHelper;

    // utility class for converting date format
    private SimpleDateFormat formatter;

    // now notification has its own class
    private NotificationHelper notificationHelper;
    private NotificationManager notificationManager;

    private final List<MqttNotif> mqttNotifications = Collections.synchronizedList(new ArrayList<MqttNotif>());
    private final List<Device> devices = Collections.synchronizedList(new ArrayList<Device>());

    private final Gson gson = new Gson();
    private Handler handler;
    private final long TWO_MINUTES = 120000;

    private String currentNotificationStatus = "";

    private BroadcastReceiver receiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            String action = intent.getAction();
            if (action == null) return;

            if (action.equals(IntentKey.WORKER_FAILED)) {
                updateNotificationContent("Workers failed");
                restartWorker(false);
            }

            if (action.equals(IntentKey.FILTER_WORKER_COMPLETED)) {
                updateNotificationContent("Workers completed");
                Bundle extras = intent.getExtras();
                if (extras != null) {

                    String deviceUuid = extras.getString(GetDevicesWorker.KEY);
                    String analyticUuid = extras.getString(GetAnalyticsWorker.KEY);
                    String resourceUuid = extras.getString(GetAnalyticsResourcesWorker.KEY);

                    runMqtt(deviceUuid, analyticUuid, resourceUuid);
                }
            }

            if (action.equals(IntentKey.FILTER_DEL_ANALYTIC)) {
                Bundle extras = intent.getExtras();
                if (extras != null) {
                    String data = extras.getString("data");
                    if (data != null) {
                        Analytic analytic = gson.fromJson(data, Analytic.class);
                        if (analytic != null) {
                            removeAnalytic(analytic);
                        }
                    }
                }
            }

            if (action.equals(IntentKey.FILTER_ADD_DEVICE)) {
                Bundle extras = intent.getExtras();
                if (extras != null) {
                    String data = extras.getString("data");

                    if (data != null) {
                        Device device = gson.fromJson(data, Device.class);
                        synchronized (devices) {
                            devices.add(device);
                            if (device != null) {
                                addNewTopicThenSubscribe(device.getDeveloperId() + "/" + device.getSn() + "/$state");
                            }
                        }

                        updateNotificationContent(currentNotificationStatus);
                    }
                }

            }

            if (action.equals(IntentKey.FILTER_DEL_DEVICE)) {
                Bundle extras = intent.getExtras();
                if (extras != null) {
                    String data = extras.getString("data");

                    if (data != null) {
                        Device device = gson.fromJson(data, Device.class);
                        if (device != null) {
                            removeDevice(device);

                            unsubscribePrefixTopic(device.getDeveloperId() + "/" + device.getSn() + "/");
                        }

                        updateNotificationContent(currentNotificationStatus);
                    }
                }
            }

            if (action.equals(IntentKey.FILTER_UPD_DEVICE)) {
                Bundle extras = intent.getExtras();
                if (extras != null) {
                    String data = extras.getString("data");
                    if (data != null) {
                        Device device = gson.fromJson(data, Device.class);
                        updateDevice(device);
                    }
                }
            }
        }
    };

    // only topic with developerId/SN/
    private void unsubscribePrefixTopic(String topic) {

        int secondIndex = topic.indexOf("/", topic.indexOf("/") + 1);
        String prefixTopic = topic.substring(0, secondIndex + 1);

        synchronized (subscribeObjList) {
            Iterator itr = subscribeObjList.iterator();
            while (itr.hasNext()) {
                final SubscribeObj obj = (SubscribeObj) itr.next();
                if (obj.topic.equals(prefixTopic)) {
                    try {
                        IMqttToken unsubToken = mqttAndroidClient.unsubscribe(obj.topic);
                        unsubToken.setActionCallback(new IMqttActionListener() {
                            @Override
                            public void onSuccess(IMqttToken asyncActionToken) {
                                Log.e(TAG, "Successfully unsubscribe : " + obj.topic);
                            }

                            @Override
                            public void onFailure(IMqttToken asyncActionToken, Throwable exception) {
                                Log.e(TAG, "Failed to unsubscribe : " + obj.topic);
                                if (exception != null && exception.getMessage() != null && exception.getMessage().length() > 0) {
                                    updateNotificationContent(exception.getMessage());
                                } else {
                                    restartWorker(false);
                                }
                            }
                        });
                    } catch (MqttException e) {
                        e.printStackTrace();
                        updateNotificationContent(e.getMessage());
                        restartWorker(false);
                    }

                    itr.remove();
                }
            }
        }
    }

    private void addNewTopicThenSubscribe(final String topic) {
        synchronized (subscribeObjList) {
            subscribeObjList.add(new SubscribeObj(topic, 0));
            try {
                IMqttToken subToken = mqttAndroidClient.subscribe(topic, 0);
                subToken.setActionCallback(new IMqttActionListener() {
                    @Override
                    public void onSuccess(IMqttToken asyncActionToken) {
                        Log.e(TAG, "Subscribe successfully : " + topic);
                    }

                    @Override
                    public void onFailure(IMqttToken asyncActionToken, Throwable exception) {
                        Log.e(TAG, "Failed to subscribe : " + topic);
                        if (exception != null && exception.getMessage() != null && exception.getMessage().length() > 0) {
                            updateNotificationContent(exception.getMessage());
                        } else {
                            restartWorker(false);
                        }

                    }
                });
            } catch (MqttException e) {
                e.printStackTrace();

                updateNotificationContent(e.getMessage());
                restartWorker(false);
            }
        }
    }

    private void initiateMqttProperties() {
        notificationId = prefs.getInt(IntentKey.KEY_NOTIF_ID, 90909090);
        brokerAddress = prefs.getString(IntentKey.KEY_MQTT_BROKER_ADDRESS, "no-host");
        brokerProtocol = prefs.getString(IntentKey.KEY_MQTT_BROKER_PROTOCOL, "ssl");
        brokerPort = prefs.getInt(IntentKey.KEY_MQTT_BROKER_PORT, 8883);
        cafile = prefs.getString(IntentKey.KEY_CAFILE, "no-cafile");
        broker = brokerProtocol + "://" + brokerAddress + ":" + brokerPort;
        username = EncryptUtil.getInstance(this).decrypt(prefs.getString(IntentKey.KEY_USER_ACCESS_TOKEN, "invalid-token"));
        password = "jwt";
        mqttClientIdentifier = prefs.getString(IntentKey.KEY_MQTT_IDENTIFIER, UUID.randomUUID().toString());
    }

    private boolean isConsecutive(Date date) {
        if (date == null) return false;

        final long timeLimit = 0;
        long secondsBetween = (Calendar.getInstance().getTime().getTime() - date.getTime()) / 1000;

        return secondsBetween <= timeLimit;
    }

    private void initateMqttClient() {
        mqttAndroidClient = new MqttAndroidClient(this, broker, mqttClientIdentifier);

        // ini kalau dari airplane
        mqttAndroidClient.setCallback(new MqttCallbackExtended() {
            @Override
            public void connectComplete(boolean reconnect, String serverURI) {
                updateNotificationContent("Connected");
                subscribe();
            }

            @Override
            public void connectionLost(Throwable exception) {
                if (exception != null && exception.getMessage() != null) {
                    boolean restartWorker = false;
                    boolean needRefresh = false;

                    if (exception instanceof MqttException) {
                        MqttException mqttException = ((MqttException) exception);
                        int reasonCode = mqttException.getReasonCode();

                        if (reasonCode == MqttException.REASON_CODE_WRITE_TIMEOUT
                                || reasonCode == MqttException.REASON_CODE_CLIENT_TIMEOUT) {
                            restartWorker = true;
                        } else if (reasonCode == MqttException.REASON_CODE_NOT_AUTHORIZED
                                || reasonCode == MqttException.REASON_CODE_FAILED_AUTHENTICATION
                                || reasonCode == MqttException.REASON_CODE_UNEXPECTED_ERROR) {
                            restartWorker = true;
                            needRefresh = true;
                        } else {

                            // jika cuma lost
                            long currentDate = new Date().getTime();
                            long issuedAt = prefs.getLong(IntentKey.KEY_TOKEN_ISSUED_AT, -1);
                            long expiredAt = prefs.getLong(IntentKey.KEY_TOKEN_EXPIRED_AT, -1);

                            if (issuedAt != -1 && expiredAt != -1) {
                                needRefresh = (expiredAt - currentDate) <= 0;
                                restartWorker = needRefresh;
                            }
                        }
                    }

                    if (restartWorker) {
                        restartWorker(needRefresh);
                    } else {
                        updateNotificationContent(exception.getMessage());
                        // jika setelah 2 menit tidak connected paksa restart
                        handler.postDelayed(new Runnable() {
                            @Override
                            public void run() {
                                if (!isClientConnected()) {
                                    restartWorker(false);
                                }
                            }
                        }, TWO_MINUTES);
                    }
                } else {
                    restartWorker(false);
                }
            }

            @Override
            public void messageArrived(String topic, MqttMessage message) {
                synchronized (mqttNotifications) {
                    for (MqttNotif notification : mqttNotifications) {
                        if (notification.getTopic().equals(topic)) {
                            if (!isConsecutive(notification.getLastMessageReceived())) {
                                String msg = new String(message.getPayload());
                                onMessageReceived(notification, getNewPayload(msg), message.isRetained());
                            }
                        }
                    }
                }

                synchronized (devices) {
                    for (Device device : devices) {
                        if (topic.equals(device.getDeveloperId() + "/" + device.getSn() + "/$state")) {
                            String msg = new String(message.getPayload());
                            onStateChanges(device, msg, message.isRetained());
                        }
                    }
                }

            }

            @Override
            public void deliveryComplete(IMqttDeliveryToken token) {
            }
        });

    }

    private void runMqtt(final String deviceUuid, final String analyticUuid, final String resourceUuid) {
        new Thread(new Runnable() {
            @Override
            public void run() {
                try {
                    updateNotificationContent("Parsing data");
                    File path = getFilesDir();
                    String devicesJson = FileUtil.readFromFile(new File(path, deviceUuid + ".json"));

                    DeviceResult deviceResult = gson.fromJson(devicesJson, DeviceResult.class);
                    List<Device> deviceList = new ArrayList<>();
                    if (deviceResult != null && deviceResult.getBody() != null) {
                        deviceList.addAll(deviceResult.getBody());
                    }
                    synchronized (devices) {
                        devices.clear();
                        devices.addAll(deviceList);
                    }

                    String analyticsJson = FileUtil.readFromFile(new File(path, analyticUuid + ".json"));
                    AnalyticResult analyticResult = gson.fromJson(analyticsJson, AnalyticResult.class);

                    List<Analytic> analyticList = new ArrayList<>();
                    if (analyticResult != null && analyticResult.getBody() != null) {
                        analyticList.addAll(analyticResult.getBody());
                    }

                    String resourcesJson = FileUtil.readFromFile(new File(path, resourceUuid + ".json"));
                    AnalyticsResourcesResult analyticsResourcesResult = gson.fromJson(resourcesJson, AnalyticsResourcesResult.class);

                    List<MqttNotif> newList = new ArrayList<>();

                    List<AnalyticResource> analyticResourceList = new ArrayList<>();
                    if (analyticsResourcesResult != null && analyticsResourcesResult.getBody() != null) {
                        analyticResourceList.addAll(analyticsResourcesResult.getBody());
                    }

                    for (AnalyticResource parent : analyticResourceList) {
                        List<Resource> resources = parent.getResources();
                        if (resources == null) continue;

                        for (Resource child : resources) {
                            Analytic analytic = getAnalytic(parent.getAnalyticId(), analyticList);

                            Device device = getDevice(child.getDeviceId(), deviceList);
                            if (analytic == null || device == null) continue;

                            MqttNotif mqttNotification = new MqttNotif(analytic, child, device);
                            newList.add(mqttNotification);
                        }
                    }


                    if (newList != null) {
                        synchronized (mqttNotifications) {
                            mqttNotifications.clear();
                            mqttNotifications.addAll(newList);
                        }
                    }

                    initSubscribeObjectList();
                    if (isClientConnected()) {
                        subscribe();
                    } else {
                        connect();
                    }

                    File dir = getFilesDir();
                    File fileDevices = new File(dir, deviceUuid + ".json");
                    boolean deleteDevice = fileDevices.delete();

                    File fileAnalytics = new File(dir, analyticUuid + ".json");
                    boolean deleteAnalytic = fileAnalytics.delete();

                    File fileResources = new File(dir, resourceUuid + ".json");
                    boolean deleteResource = fileResources.delete();

                    Log.e(TAG, "delete " + deleteDevice + " " + deleteAnalytic + " " + deleteResource);

                } catch (IllegalStateException | JsonSyntaxException exception) {
                    updateNotificationContent(exception.getMessage());
                    restartWorker(false);
                    Log.e(TAG, exception.getMessage());
                }
            }
        }).start();
    }

    private boolean isTopicExists(String topic, List<SubscribeObj> currentList) {
        for (int i = 0; i < currentList.size(); i++) {
            if (currentList.get(i).topic.equals(topic)) {
                return true;
            }
        }

        return false;
    }

    private void initSubscribeObjectList() {
        int defaultQos = 0;
        List<SubscribeObj> newSubscribeList = new ArrayList<>();

        synchronized (mqttNotifications) {
            for (MqttNotif notification : mqttNotifications) {

                if (!isTopicExists(notification.getTopic(), newSubscribeList)) {
                    newSubscribeList.add(new SubscribeObj(notification.getTopic(), defaultQos));
                }
            }
        }

        synchronized (devices) {

            for (Device device : devices) {
                // ${widget.device.developerId}/${widget.device.sn}/\$state
                String deviceTopic = device.getDeveloperId() + "/" + device.getSn() + "/$state";
                if (!isTopicExists(deviceTopic, newSubscribeList)) {
                    newSubscribeList.add(new SubscribeObj(deviceTopic, defaultQos));
                }
            }
        }

        if (newSubscribeList != null) {
            synchronized (subscribeObjList) {
                subscribeObjList.clear();
                subscribeObjList.addAll(newSubscribeList);
            }
        }
    }

    private void updateDevice(Device device) {
        synchronized (devices) {
            for (Device item : devices) {
                if (item.getId().equals(device.getId())) {
                    item.setName(device.getName());
                    item.setDesc(device.getDesc());
                    item.setOptions(device.getOptions());
                    break;
                }
            }
        }
    }

    private void removeDevice(Device device) {
        synchronized (devices) {
            Iterator itr = devices.iterator();
            while (itr.hasNext()) {
                Device item = (Device) itr.next();
                if (item.getId().equals(device.getId())) {
                    itr.remove();
                    break;
                }
            }
        }
    }

    private void removeAnalytic(Analytic analytic) {
        String removedTopic = null;

        synchronized (mqttNotifications) {
            Iterator itr = mqttNotifications.iterator();
            while (itr.hasNext()) {
                final MqttNotif notification = (MqttNotif) itr.next();
                if (notification.getAnalytic().getId().equals(analytic.getId())) {
                    removedTopic = notification.getTopic();
                    try {
                        IMqttToken unsubToken = mqttAndroidClient.unsubscribe(notification.getTopic());
                        unsubToken.setActionCallback(new IMqttActionListener() {
                            @Override
                            public void onSuccess(IMqttToken asyncActionToken) {
                                Log.e(TAG, "Successfully unsubscribe : " + notification.getTopic());
                            }

                            @Override
                            public void onFailure(IMqttToken asyncActionToken, Throwable exception) {
                                Log.e(TAG, "Failed to unsubscribe : " + notification.getTopic());
                                if (exception != null && exception.getMessage() != null && exception.getMessage().length() > 0) {
                                    updateNotificationContent(exception.getMessage());
                                } else {
                                    restartWorker(false);
                                }

                            }
                        });
                    } catch (MqttException e) {
                        e.printStackTrace();
                        updateNotificationContent(e.getMessage());
                        restartWorker(false);
                    }

                    itr.remove();
                    break;
                }
            }
        }

        if (removedTopic != null) {
            synchronized (subscribeObjList) {
                Iterator itr = subscribeObjList.iterator();
                while (itr.hasNext()) {
                    SubscribeObj obj = (SubscribeObj) itr.next();
                    if (obj.topic.equals(removedTopic)) {
                        itr.remove();
                        break;
                    }
                }

            }
        }
    }

    private Analytic getAnalytic(String analyticId, List<Analytic> analytics) {
        for (Analytic a : analytics) {
            if (a.getId().equals(analyticId)) return a;
        }
        return null;
    }

    private Device getDevice(String deviceId, List<Device> list) {
        for (Device a : list) {
            if (a.getId().equals(deviceId)) return a;
        }
        return null;
    }

    private void initTopics() {
        updateNotificationContent("Retrieving data");
        WorkManager.getInstance(getApplicationContext()).cancelUniqueWork("requestDataForService");

        final OneTimeWorkRequest getDevicesRequest = new OneTimeWorkRequest.Builder(GetDevicesWorker.class)
                .build();
        final OneTimeWorkRequest getAnalyticsRequest = new OneTimeWorkRequest.Builder(GetAnalyticsWorker.class)
                .build();

        final Data data = new Data.Builder()
                .putString(IntentKey.KEY_BROADCAST, IntentKey.FILTER_WORKER_COMPLETED)
                .build();
        final OneTimeWorkRequest getAnalyticsResourcesRequest = new OneTimeWorkRequest.Builder(GetAnalyticsResourcesWorker.class)
                .setInputData(data)
                .build();

        WorkManager.getInstance(getApplicationContext())
                .beginUniqueWork("requestDataForService", ExistingWorkPolicy.REPLACE, getDevicesRequest)
                .then(getAnalyticsRequest)
                .then(getAnalyticsResourcesRequest)
                .enqueue();
    }

    private static class SubscribeObj {
        String topic;
        int qos;

        SubscribeObj(String topic, int qos) {
            this.topic = topic;
            this.qos = qos;
        }
    }

    // subscribe to broker
    private void subscribe() {
        if (subscribeObjList.size() > 0) {
            final String[] topics = new String[subscribeObjList.size()];
            int[] qoss = new int[subscribeObjList.size()];

            for (int i = 0; i < subscribeObjList.size(); i++) {
                topics[i] = subscribeObjList.get(i).topic;
                qoss[i] = subscribeObjList.get(i).qos;
            }

            updateNotificationContent("Subscribing");
            try {
                IMqttToken subToken = mqttAndroidClient.subscribe(topics, qoss);
                subToken.setActionCallback(new IMqttActionListener() {
                    @Override
                    public void onSuccess(IMqttToken asyncActionToken) {
                        updateNotificationContent("Connected");
                    }

                    @Override
                    public void onFailure(IMqttToken asyncActionToken, Throwable exception) {
                        if (exception != null && exception.getMessage() != null && exception.getMessage().length() > 0) {
                            updateNotificationContent(exception.getMessage());
                        } else {
                            restartWorker(false);
                        }

                    }
                });
            } catch (MqttException e) {
                updateNotificationContent(e.getMessage());
                restartWorker(false);
                e.printStackTrace();
            }
        }
    }

    private void connect() {
        username = EncryptUtil.getInstance(this).decrypt(prefs.getString(IntentKey.KEY_USER_ACCESS_TOKEN, "invalid-token"));

        MqttConnectOptions mqttConnectOptions = new MqttConnectOptions();
        // reconnect and clean session variable for mqtt connect options
        mqttConnectOptions.setAutomaticReconnect(true);
        mqttConnectOptions.setCleanSession(true);
        mqttConnectOptions.setUserName(username);
        mqttConnectOptions.setPassword(password.toCharArray());
        mqttConnectOptions.setMqttVersion(MqttConnectOptions.MQTT_VERSION_3_1);
        mqttConnectOptions.setKeepAliveInterval(10);
        mqttConnectOptions.setConnectionTimeout(20);

        if (brokerProtocol.equals("ssl")) {
            //load certificate file if this broker connect via secure ssl
            SocketFactory.SocketFactoryOptions socketFactoryOptions = new SocketFactory.SocketFactoryOptions();
            try {
                InputStream targetStream = new ByteArrayInputStream(cafile.getBytes());
                socketFactoryOptions.withCaInputStream(targetStream);
                mqttConnectOptions.setSocketFactory(new SocketFactory(socketFactoryOptions));
            } catch (IOException | NoSuchAlgorithmException | KeyStoreException | CertificateException | KeyManagementException | UnrecoverableKeyException e) {
                e.printStackTrace();
            }
        }

        try {
            mqttAndroidClient.connect(mqttConnectOptions, null, new IMqttActionListener() {
                @Override
                public void onSuccess(IMqttToken asyncActionToken) {
                    updateNotificationContent("Connected");
                    subscribe();
                }

                @Override
                public void onFailure(IMqttToken asyncActionToken, Throwable exception) {
                    if (exception != null && exception.getMessage() != null) {

                        boolean isConnected = false;
                        if (exception instanceof MqttException) {
                            int reasonCode = ((MqttException) exception).getReasonCode();
                            if (reasonCode == MqttException.REASON_CODE_CLIENT_CONNECTED) {
                                isConnected = true;
                            }
                        }

                        if (isConnected) updateNotificationContent("Connected");
                        else updateNotificationContent(exception.getMessage());
                    } else {
                        restartWorker(false);
                    }
                }
            });
        } catch (MqttException e) {
            e.printStackTrace();

            updateNotificationContent(e.getMessage());
            restartWorker(false);
        }
    }

    private void initiateLocalVariable() {
        handler = new Handler(Looper.getMainLooper());
        prefs = PreferenceManager.getDefaultSharedPreferences(this);
        dbHelper = WidgetReaderDbHelper.getInstance(this);
        notificationHelper = new NotificationHelper(this);
        formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss", Locale.US);
        notificationManager = (NotificationManager) getSystemService(Context.NOTIFICATION_SERVICE);
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
            NotificationChannel alarmNotificationChannel = new NotificationChannel(
                    NotificationHelper.ALARM_CHANNEL_ID, NotificationHelper.ALARM_CHANNEL_NAME,
                    NotificationManager.IMPORTANCE_HIGH);
            alarmNotificationChannel.setLockscreenVisibility(NotificationCompat.VISIBILITY_PUBLIC);
            alarmNotificationChannel.setVibrationPattern(new long[]{100, 200, 300, 400, 500, 400, 300, 200, 400});
            alarmNotificationChannel.enableVibration(true);
            alarmNotificationChannel.enableLights(true);
            alarmNotificationChannel.setLightColor(Color.RED);

            notificationManager.createNotificationChannel(alarmNotificationChannel);

            NotificationChannel serverStatusNotificationChannel = new NotificationChannel(
                    NotificationHelper.SERVER_STATUS_CHANNEL_ID, NotificationHelper.SERVER_STATUS_CHANNEL_NAME,
                    NotificationManager.IMPORTANCE_DEFAULT);
            serverStatusNotificationChannel.setSound(null, null);
            notificationManager.createNotificationChannel(serverStatusNotificationChannel);
        }
    }

    private void registerLocalReceiver() {
        IntentFilter filter = new IntentFilter();
        filter.addAction(IntentKey.FILTER_WORKER_COMPLETED);
        filter.addAction(IntentKey.WORKER_FAILED);
        filter.addAction(IntentKey.FILTER_ADD_DEVICE);
        filter.addAction(IntentKey.FILTER_DEL_DEVICE);
        filter.addAction(IntentKey.FILTER_UPD_DEVICE);
        filter.addAction(IntentKey.FILTER_DEL_ANALYTIC);

        LocalBroadcastManager.getInstance(this).registerReceiver(receiver, filter);
    }

    private void restartWorker(boolean needRefresh) {
        Constraints constraints = new Constraints.Builder()
                .setRequiredNetworkType(NetworkType.CONNECTED)
                .build();

        OneTimeWorkRequest.Builder restartRequestBuilder = new OneTimeWorkRequest.Builder(RestartServiceWorker.class)
                .setConstraints(constraints);
        if (needRefresh) {
            Data data = new Data.Builder()
                    .putBoolean(IntentKey.KEY_NEED_TO_REFRESH, true)
                    .build();

            restartRequestBuilder.setInputData(data);
            updateNotificationContent("Restarting service by also updating token");
        } else {
            updateNotificationContent("Restarting service");
        }


        WorkManager.getInstance(getApplicationContext())
                .enqueueUniqueWork("requestRestartService", ExistingWorkPolicy.REPLACE, restartRequestBuilder.build());
    }

    private void unsubscribeAll() {
        synchronized (subscribeObjList) {
            String[] subscribed = new String[subscribeObjList.size()];
            int index = 0;
            for (SubscribeObj obj : subscribeObjList) {
                subscribed[index++] = obj.topic;
            }

            if (subscribed.length > 0) {
                try {
                    updateNotificationContent("Unsubscribing");
                    IMqttToken unsubAllToken = mqttAndroidClient.unsubscribe(subscribed);
                    unsubAllToken.setActionCallback(new IMqttActionListener() {
                        @Override
                        public void onSuccess(IMqttToken asyncActionToken) {
                            updateNotificationContent("Successfully unsubscribed all topics");
                            initTopics();
                        }

                        @Override
                        public void onFailure(IMqttToken asyncActionToken, Throwable exception) {
                            if (exception != null && exception.getMessage() != null && exception.getMessage().length() > 0) {
                                updateNotificationContent(exception.getMessage());
                            } else {
                                restartWorker(false);
                            }


                        }
                    });
                } catch (MqttException e) {
                    e.printStackTrace();

                    updateNotificationContent(e.getMessage());
                    restartWorker(false);
                }
            }
        }
    }

    void scheduleRefreshToken() {
        String token = EncryptUtil.getInstance(this).decrypt(prefs.getString(IntentKey.KEY_USER_ACCESS_TOKEN, "invalid-token"));
        if (token != null && !token.equals("invalid-token")) {
            WorkManager.getInstance(getApplicationContext()).cancelUniqueWork("SamServiceScheduledRefreshToken");

            Data data = new Data.Builder()
                    .putBoolean(IntentKey.KEY_NEED_TO_REFRESH, true)
                    .build();

            // start repeating check
            OneTimeWorkRequest oneTime = new OneTimeWorkRequest.Builder(RestartServiceWorker.class)
                    .setInitialDelay(12, TimeUnit.HOURS)
                    .setInputData(data)
                    .build();
            WorkManager.getInstance(getApplicationContext())
                    .beginUniqueWork("SamServiceScheduledRefreshToken", ExistingWorkPolicy.REPLACE, oneTime).enqueue();

        }
    }

    @Override
    public void onCreate() {
        super.onCreate();

        initiateLocalVariable();
        registerLocalReceiver();
        initiateMqttProperties();
        initateMqttClient();
        scheduleRefreshToken();
    }

    boolean isClientConnected() {
        if (mqttAndroidClient != null) {
            try {
                boolean result = mqttAndroidClient.isConnected();
                if (result) {
                    updateNotificationContent("Connected");
                    return true;
                } else {
                    return false;
                }
            } catch (Exception e) {
                e.printStackTrace();
                return false;
            }
        } else {
            return false;
        }
    }

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        if (username != null && !username.equals("invalid-token") && username.length() > 0) {
            if (isClientConnected()) {
                unsubscribeAll();
            } else {
                initTopics();
            }

            startForeground(notificationId, getNotificationForForegroundService("Starting"));
            return Service.START_STICKY;
        }

        stopSelf();
        return Service.START_NOT_STICKY;
    }

    @Nullable
    @Override
    public IBinder onBind(Intent intent) {
        // method not used
        return null;
    }

    @Override
    public void onDestroy() {
        if (mqttAndroidClient != null) {
            mqttAndroidClient.close();
            disconnectClient();
            mqttAndroidClient.unregisterResources();
        }

        stopForeground(true);

        LocalBroadcastManager.getInstance(this).unregisterReceiver(receiver);
        receiver = null;

        if (dbHelper != null) dbHelper.close();

        try {
            WorkManager.getInstance(this).cancelUniqueWork("SamServiceScheduledRefreshToken");
            WorkManager.getInstance(this).cancelUniqueWork("requestRestartServiceFromPlugin");
            WorkManager.getInstance(this).cancelUniqueWork("requestRestartService");
            WorkManager.getInstance(this).cancelUniqueWork("requestDataForService");
        } catch (Exception e) {
            Log.e(TAG, e.getMessage() == null ? "null" : e.getMessage());
        }

        try {
            prefs.edit().remove(IntentKey.KEY_USER_ACCESS_TOKEN).apply();

            if (notificationManager != null) {
                notificationManager.cancel(notificationId);
            }
        } catch (Exception e) {
            Log.e(TAG, e.getMessage() == null ? "null" : e.getMessage());
        }

        stopSelf();
        super.onDestroy();
    }

    private void disconnectClient() {
        if (isClientConnected()) {
            try {
                IMqttToken disToken = mqttAndroidClient.disconnect(3000);
                disToken.setActionCallback(new IMqttActionListener() {
                    @Override
                    public void onSuccess(IMqttToken asyncActionToken) {
                        Log.e(TAG, "Successfully disconnect");
                    }

                    @Override
                    public void onFailure(IMqttToken asyncActionToken, Throwable exception) {
                        Log.e(TAG, "Failed to disconnect");
                    }
                });
            } catch (MqttException e) {
                e.printStackTrace();
            }
        }
    }

    private String getDeviceOnlineStatistics() {
        int totalDevices = devices.size();
        int totalOnline = 0;
        int totalLost = 0;
        for (Device device : devices) {
            if (device.getState() != null && device.getState().equalsIgnoreCase("ready")) {
                totalOnline++;
            } else {
                totalLost++;
            }
        }

        if (totalOnline == 0) return "<br>No active devices";

        return "<br>"+totalOnline+" active devices out of a total of " + totalDevices;

        // return "<br>Total : " + totalDevices + " devices (status &#127383; " + totalOnline + " &#10062; " + totalLost + ")";
    }

    private Notification getNotificationForForegroundService(String status) {
        Intent notificationIntent = new Intent(this, FlutterUtil.getMainActivityClass(this));
        PendingIntent pendingIntent = PendingIntent.getActivity(this, 0,
                notificationIntent, 0);

        Intent restartIntent = new Intent(this, RestartServiceReceiver.class);
        restartIntent.putExtra(IntentKey.KEY_NOTIF_ID, notificationId);
        restartIntent.setAction(IntentKey.RESTART_ACTION);
        PendingIntent piRestart = PendingIntent.getBroadcast(this, 0, restartIntent, 0);

        String sampleText = "<b>" + status + "</b>";
        sampleText += getDeviceOnlineStatistics();
        sampleText += "<br><br><i>To hide this notification, go to application notification settings and uncheck \"Show notifications\"</i>";

        Spanned styledText = Html.fromHtml(sampleText);

        NotificationCompat.Builder builder = new NotificationCompat.Builder(this,
                NotificationHelper.SERVER_STATUS_CHANNEL_ID)
                .setContentTitle("Notification service")
                .setStyle(new NotificationCompat.BigTextStyle()
                        .bigText(styledText))
                .setContentText(status)
                .setCategory(NotificationCompat.CATEGORY_SERVICE)
                .addAction(R.drawable.ic_restart, "Restart", piRestart)
                .setContentIntent(pendingIntent)
                .setOngoing(true);

//        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
//            Intent settingsIntent = new Intent(this, NotificationBroadcastReceiver.class);
//            PendingIntent piSettings = PendingIntent.getBroadcast(this, 0, settingsIntent, 0);
//
//            builder.addAction(R.drawable.ic_cog_outline, "Settings", piSettings);
//        }

        // set notification logo
        Bitmap icon = StateIndicatorHelper.decodeSampledBitmapFromResource(getResources(), R.mipmap.ic_launcher, 100, 100);
        builder.setSmallIcon(R.drawable.ic_link);
        builder.setLargeIcon(icon);

        return builder.build();

    }

    private void updateNotificationContent(String status) {
        currentNotificationStatus = status;

        NotificationManager notificationManager = (NotificationManager)
                getSystemService(Context.NOTIFICATION_SERVICE);
        if (notificationManager != null) {
            notificationManager.notify(notificationId, getNotificationForForegroundService(status));
        }
    }

    // splitting string resulted from mqtt received
    // for value result, contains datetime. (<datetime>;<value>)
    // for boolean result, only value (true or false)
    private MqttPayload getNewPayload(String fullMessage) {

        Date dateTime = Calendar.getInstance().getTime();
        String valueMessage = fullMessage;
        String extraMessage = null;

        if (fullMessage.contains(";")) {
            try {
                String[] separatedMessage = fullMessage.split(";");
                if (separatedMessage.length >= 2) {

                    dateTime = formatter.parse(separatedMessage[0]);
                    valueMessage = separatedMessage[1];
                }

            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        if (valueMessage.contains("|")) {
            String[] splitted = valueMessage.split("\\|");

            if (splitted.length >= 2) {
                valueMessage = splitted[0];
                extraMessage = splitted[1];
            }
        }

        return new MqttPayload(dateTime, valueMessage, extraMessage);
    }

    private boolean isStateChecked(String checkedString, String state) {
        if (checkedString.isEmpty()) return false;

        return checkedString.contains(state.substring(0, 1));
    }

    private void onStateChanges(Device device, String state, boolean isRetained) {
        device.setState(state);
        updateNotificationContent(currentNotificationStatus);

        if (device.getOptions() == null) {
            return;
        }

        if (device.getOptions().getNotify() && !state.equals(device.getState())
                && isStateChecked(device.getOptions().getNotify_checked(), state) && !isRetained) {
            showDeviceNotification(device, state);
        }

        if (device.getOptions().getAlert() && !state.equals(device.getState())
                && isStateChecked(device.getOptions().getAlert_checked(), state) && !isRetained) {
            launchAlertActivity(device, state);
        }

    }

    private Class<?> getProviderClass(String type) {
        switch (type) {
            case AppWidgetType.WGT_MONITOR:
                return SimpleTextWidget.class;
            case AppWidgetType.WGT_LINE_CHART:
                return DemoAppWidget.class;
            default:
                return null;
        }
    }

    private void updateAppWidgetContent(String topic, String value) {
        if (value == null || value.length() == 0) return;

        List<AppWidgetObject> allList = dbHelper.getAllWhereTopic(topic);

        if (allList.size() > 0) {
            boolean isConsecutive = false;
            try {
                Date lastDate = formatter.parse(allList.get(0).getDatetime());
                Date currentDate = new Date();

                long diff = (currentDate.getTime() - lastDate.getTime()) / 1000;

                if (diff < 2) isConsecutive = true;
            } catch (ParseException e) {
                e.printStackTrace();
            }

            if (isConsecutive) return;

            List<AppWidgetObject> list = dbHelper.updateValueWhereTopic(allList, topic, value);

            HashMap<String, List<Integer>> map = new HashMap<>();

            for (AppWidgetObject obj : list) {
                if (!map.containsKey(obj.getType())) {
                    map.put(obj.getType(), new ArrayList<Integer>());
                }

                map.get(obj.getType()).add(Integer.parseInt(obj.getWidgetId()));
            }

            for (String key : map.keySet()) {
                int[] ids = new int[map.get(key).size()];
                int index = 0;
                for (Integer id : map.get(key)) {
                    ids[index] = id.intValue();
                    index++;
                }

                Class providerClass = getProviderClass(key);
                if (providerClass != null) {
                    Intent intent = new Intent(this, providerClass);
                    intent.setAction(AppWidgetManager.ACTION_APPWIDGET_UPDATE);
                    intent.putExtra(AppWidgetManager.EXTRA_APPWIDGET_IDS, ids);
                    sendBroadcast(intent);
                }
            }
        }
    }

    private void onMessageReceived(MqttNotif notification, MqttPayload payload, boolean isRetained) {
        try {
            if (payload == null) return;

            Log.d(TAG, "notification " + notification.getResource().getOptions().getRule()
                    + " " + notification.getResource().getOptions().getThreshold());

            Log.d(TAG, "Incoming message: " + payload.getValue());

            if (!isRetained) {
                updateAppWidgetContent(notification.getTopic(), payload.getValue());
            }

            boolean isExceedThreshold = exceedThreshold(
                    payload.getValue(),
                    notification);

            Log.d(TAG, "Exceed threshold: " + isExceedThreshold);

            if (isExceedThreshold) {
                boolean isNotify = notification.getResource().getOptions().isNotify();
                boolean isAlert = notification.getResource().getOptions().isAlert();

                Log.d(TAG, "IsNotify: " + isNotify + " IsAlert: " + isAlert);

                if (isNotify && !isRetained) {
                    System.out.println("show notification");
                    showNotification(notification, payload.getValue());
                }

                if (isAlert && !isRetained) {
                    System.out.println("Show alarm");
                    launchAlarmActivity(notification, payload.getValue());
                }

                if (isNotify || isAlert) {
                    // update last message received
                    notification.setLastMessageReceived(Calendar.getInstance().getTime());
                }
            }

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private boolean xnor(boolean a, boolean b) {
        return a == b;
    }

    private boolean exceedThreshold(String value, MqttNotif notification) {
        String rule = notification.getResource().getOptions().getRule();
        double threshold = notification.getResource().getOptions().getThreshold();

        if (rule.equals("#")) {
            // boolean datatype
            boolean isThresholdPassed = Boolean.parseBoolean(value) == (threshold == 1);

            // more checking on activeState
            return xnor(isThresholdPassed, notification.isActiveState());
        }

        if (value != null && value.length() > 0) {
            try {
                switch (rule) {
                    case ">":
                        return Double.parseDouble(value) > threshold;
                    case "<":
                        return Double.parseDouble(value) < threshold;
                    case "=":
                        return Double.parseDouble(value) == threshold;
                    default:
                        return false;
                }
            } catch (NumberFormatException e) {
                Log.e(TAG, "Number format not valid " + value);
            }
        }

        return false;
    }

    private String getComparisonString(String comp) {
        switch (comp) {
            case ">":
                return "bigger than";
            case "<":
                return "lower than";
            case "=":
                return "equals";
        }

        return "";
    }

    // build message for notification message on notification bar/action bar
    private String buildMessage(MqttNotif notification, String value) {
        if (notification != null) {
            // if boolean
            if (notification.getResource().getOptions().getRule().equals("#")) {
                return value.toUpperCase() + notification.getUnit();
            } else {
                return value.toUpperCase() + notification.getUnit() + " (" + getComparisonString(notification.getResource().getOptions().getRule()) + " "
                        + notification.getResource().getOptions().getThreshold() + notification.getUnit() + ")";
            }
        }
        return "";
    }

    private void launchAlertActivity(Device device, String state) {
        Intent fullScreenIntent = new Intent(this, AlertActivity.class);
        String message = StateIndicatorHelper.getStateName(state);
        if (device.getOptions() != null && device.getOptions().getAlarmmsg() != null) {
            String msgByState = device.getOptions().getAlarmmsg().getMsgByState(state);
            if (msgByState != null && msgByState.length() > 0) {
                message = msgByState;
            }
        }

        Bundle extras = new Bundle();
        extras.putString(IntentKey.EXTRA_VALUE, message);
        extras.putString(IntentKey.EXTRA_STATE, state);
        extras.putString(IntentKey.EXTRA_TITLE, device.getName());
        extras.putString(IntentKey.EXTRA_DEVICE, device.getSn());

        final int notificationAlertId = Integer.parseInt(device.getId());
        extras.putInt(IntentKey.KEY_NOTIF_ID, notificationAlertId);

        fullScreenIntent.putExtras(extras);
        fullScreenIntent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK |
                Intent.FLAG_ACTIVITY_NO_HISTORY | Intent.FLAG_ACTIVITY_EXCLUDE_FROM_RECENTS);

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q) {
            Bitmap largeIcon = StateIndicatorHelper.getStateImage(getResources(), state);

            PendingIntent fullScreenPendingIntent = PendingIntent.getActivity(this, 0,
                    fullScreenIntent, 0);

            Uri soundUri = RingtoneManager.getDefaultUri(RingtoneManager.TYPE_ALARM);

            Intent dismissIntent = new Intent(this, AlertCancelReceiver.class);
            dismissIntent.setAction(IntentKey.DISMISS_ACTION);
            dismissIntent.putExtra(IntentKey.KEY_NOTIF_ID, notificationAlertId);
            PendingIntent piDismiss = PendingIntent.getBroadcast(this, 0, dismissIntent, 0);

            // create builder
            NotificationCompat.Builder notificationBuilder =
                    new NotificationCompat.Builder(this, NotificationHelper.ALARM_CHANNEL_ID)
                            .setSmallIcon(R.drawable.ic_sam_notification_logo_trans)
                            .setContentTitle("ALERT for " + device.getName() + " (" + device.getSn() + ")")
                            .setContentText("The current device state is " + state)
                            .setSound(soundUri)
                            .setPriority(NotificationCompat.PRIORITY_HIGH)
                            .setCategory(NotificationCompat.CATEGORY_ALARM)
                            .setLargeIcon(largeIcon)
                            .setAutoCancel(true)
                            .setOngoing(true)
                            .setVibrate(new long[]{100, 200, 300, 400, 500, 400, 300, 200, 400})
                            .addAction(R.drawable.ic_close, "Dismiss", piDismiss)
                            .setFullScreenIntent(fullScreenPendingIntent, true);

            Notification n = notificationBuilder.build();
            n.flags |= Notification.FLAG_INSISTENT;

            if (notificationManager != null) {
                notificationManager.notify(notificationAlertId, n);
                new Handler().postDelayed(new Runnable() {
                    @Override
                    public void run() {
                        notificationManager.cancel(notificationAlertId);
                    }
                }, IntentKey.TIME_OUT);
            }
        } else {
            startActivity(fullScreenIntent);
        }
    }

    private void launchAlarmActivity(MqttNotif notification, String valueReceived) {
        Intent fullScreenIntent = new Intent(this, AlarmActivity.class);
        Bundle extras = new Bundle();

        valueReceived = valueReceived + notification.getUnit();
        if (notification.getResource().getOptions().getText() != null &&
                notification.getResource().getOptions().getText().length() > 0) {
            valueReceived = notification.getResource().getOptions().getText();
        }
        extras.putString(IntentKey.EXTRA_VALUE, valueReceived);
        extras.putString(IntentKey.EXTRA_TITLE, notification.getAnalytic().getTitle());
        extras.putString(IntentKey.EXTRA_DEVICE, notification.getDevice().getName());

        final int notificationAlarmId = Integer.parseInt(notification.getAnalytic().getId());
        extras.putInt(IntentKey.KEY_NOTIF_ID, notificationAlarmId);

        fullScreenIntent.putExtras(extras);
        fullScreenIntent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK |
                Intent.FLAG_ACTIVITY_NO_HISTORY | Intent.FLAG_ACTIVITY_EXCLUDE_FROM_RECENTS);

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.Q) {
            // set large icon for notification
            Bitmap largeIcon = AnalyticIconHelper.getLargeIconBitmap(getResources(), notification.getAnalytic().getModel());

            PendingIntent fullScreenPendingIntent = PendingIntent.getActivity(this, 0,
                    fullScreenIntent, 0);

            Uri soundUri = RingtoneManager.getDefaultUri(RingtoneManager.TYPE_ALARM);

            Intent dismissIntent = new Intent(this, AlertCancelReceiver.class);
            dismissIntent.setAction(IntentKey.DISMISS_ACTION);
            dismissIntent.putExtra(IntentKey.KEY_NOTIF_ID, notificationAlarmId);
            PendingIntent piDismiss = PendingIntent.getBroadcast(this, 0, dismissIntent, 0);

            // create builder
            NotificationCompat.Builder notificationBuilder =
                    new NotificationCompat.Builder(this, NotificationHelper.ALARM_CHANNEL_ID)
                            .setSmallIcon(R.drawable.ic_sam_notification_logo_trans)
                            .setContentTitle("ALERT for " + notification.getAnalytic().getTitle())
                            .setContentText(notification.getDevice().getName() + " reaches a value of " + valueReceived + notification.getUnit())
                            .setSound(soundUri)
                            .setPriority(NotificationCompat.PRIORITY_HIGH)
                            .setCategory(NotificationCompat.CATEGORY_ALARM)
                            .setLargeIcon(largeIcon)
                            .setAutoCancel(true)
                            .setOngoing(true)
                            .addAction(R.drawable.ic_close, "Dismiss", piDismiss)
                            .setVibrate(new long[]{100, 200, 300, 400, 500, 400, 300, 200, 400})
                            .setFullScreenIntent(fullScreenPendingIntent, true);

            Notification n = notificationBuilder.build();
            n.flags |= Notification.FLAG_INSISTENT;

            if (notificationManager != null) {
                notificationManager.notify(notificationAlarmId, n);
                new Handler().postDelayed(new Runnable() {
                    @Override
                    public void run() {
                        notificationManager.cancel(notificationAlarmId);
                    }
                }, IntentKey.TIME_OUT);
            }
        } else {
            startActivity(fullScreenIntent);
        }
    }

    private void showDeviceNotification(Device device, String state) {
        Log.d(TAG, "Trying to show device notification");
        String message = device.getName() + " is " + StateIndicatorHelper.getStateName(state);
        if (device.getOptions() != null && device.getOptions().getNotifymsg() != null) {
            String msgByState = device.getOptions().getNotifymsg().getMsgByState(state);
            if (msgByState != null && msgByState.length() > 0) {
                message = msgByState;
            }
        }

        String notificationTitle = device.getName();
        Bitmap largeIcon = StateIndicatorHelper.getStateImage(getResources(), state);

        notificationHelper.createNotification("999" + device.getId(),
                notificationTitle, message, largeIcon);
        Log.d(TAG, "End notification");
    }

    private void showNotification(MqttNotif notification, String valueReceived) {
        Log.d(TAG, "Trying to show notification");
        String message = buildMessage(notification, valueReceived);
        if (notification.getResource().getOptions().getText() != null && notification.getResource().getOptions().getText().length() > 0) {
            message = notification.getResource().getOptions().getText();
        }

        String notificationTitle = notification.getAnalytic().getTitle();
        Bitmap largeIcon = AnalyticIconHelper.getLargeIconBitmap(getResources(), notification.getAnalytic().getModel());

        notificationHelper.createNotification("111" + notification.getAnalytic().getId(), notificationTitle, message, largeIcon);
        Log.d(TAG, "End notification");
    }

}